/**
 * This program has been developed by students from the bachelor Computer Science at
 * Utrecht University within the Software Project course.
 * © Copyright Utrecht University (Department of Information and Computing Sciences)
 */
import { Edge as ReactEdge, NodeProps } from 'reactflow';

/** Enums for the possible types of query elements */
export enum QueryElementTypes {
  Entity = 'entity',
  Relation = 'relation',
  Attribute = 'attribute',
  Function = 'function',
}

/** List of possible query element types */
export const possibleTypes: string[] = [
  QueryElementTypes.Entity,
  QueryElementTypes.Relation,
  QueryElementTypes.Attribute,
  QueryElementTypes.Function,
];

/** All the possible react flow nodes. */
export type AnyNode = EntityNode | RelationNode | AttributeNode | FunctionNode;
export type Edge = ReactEdge<any>;
export type AnyElement = AnyNode | Edge;
export type AnyNodeData =
  | EntityData
  | RelationData
  | AttributeData
  | FunctionData;

export type EntityNode = NodeProps<EntityData>;
export type RelationNode = NodeProps<RelationData>;
export type AttributeNode = NodeProps<AttributeData>;
export type FunctionNode = NodeProps<FunctionData>;
export type ModifierNode = NodeProps<ModifierData>;

export interface NodeData {
  fadeIn: boolean;
}

export interface ModifierData {
  type: string;
}

/** Interface for the data in an entity node. */
export interface EntityData extends NodeData {
  name: string;
}

/** Interface for the data in an relation node. */
export interface RelationData extends NodeData {
  name: string;
  collection: string;
  depth: { min: number; max: number };
}

/** Interface for the data in an attribute node.
 * Can have multiple constraint datatypes.
 * string MatchTypes: exact/contains/startswith/endswith.
 * int    MatchTypes: GT/LT/EQ.
 * bool   MatchTypes: EQ/NEQ.
 */
export interface AttributeData extends NodeData {
  name: string;
  value: string;
  dataType: string;
  matchType: string;
}

export interface FunctionArg {
  displayName: string;
  connectable: boolean; // does this arg have a connectable handle?
  value: string | undefined; // undefined means no text input
  visible: boolean; // for implicit connections
}

export interface FunctionArgs {
  [name: string]: FunctionArg;
}

/** Interface for the data in a function node. */
export interface FunctionData extends NodeData {
  functionType: string;
  args: FunctionArgs;
}

/** Interface for updating the edges. */
export interface updateEdges {
  newEdge: Edge | undefined;
  removeEdge: Edge | undefined;
}

/**
 * Checks if a node is an entityNode.
 * @param {AnyNode} node The node that has to checked.
 * @returns True and casts if the node is an EntityNode.
 */
export function isEntityNode(node: AnyNode): node is EntityNode {
  return node.type == QueryElementTypes.Entity;
}

/**
 * Checks if a node is a RelationNode.
 * @param {AnyNode} node The node that has to checked.
 * @returns True and casts if the node is an RelationNode.
 */
export function isRelationNode(node: AnyNode): node is RelationNode {
  return node.type == QueryElementTypes.Relation;
}

/**
 * Checks if a node is an AttributeNode.
 * @param {AnyNode} node The node that has to checked.
 * @returns True and casts if the node is an AttributeNode.
 */
export function isAttributeNode(node: AnyNode): node is AttributeNode {
  return node.type == QueryElementTypes.Attribute;
}

/**
 * Checks if a node is an FunctionNode.
 * @param {AnyNode} node The node that has to checked.
 * @returns True and casts if the node is an FunctionNode.
 */
export function isFunctionNode(node: AnyNode): node is FunctionNode {
  return node.type == QueryElementTypes.Function;
}
